"""
pydantic xml serialization/deserialization extension
"""

from . import config, errors
from .errors import ModelError
from .model import BaseGenericXmlModel, BaseXmlModel, XmlAttributeInfo, XmlElementInfo, XmlWrapperInfo, attr, element
from .model import wrapped
from .serializers import DEFAULT_ENCODER, XmlEncoder



import typing
import xml.etree.ElementTree as std_etree
from typing import Any, Dict, Optional

from lxml import etree

__all__ = (
    'etree',
    'create_element',
)


def create_element(
        tag: str,
        attrib: Optional[Dict[str, Any]] = None,
        nsmap: Optional[Dict[str, str]] = None,
) -> std_etree.Element:
    element = etree.Element(
        tag,
        attrib=attrib,
        # https://github.com/lxml/lxml-stubs/issues/76
        nsmap={ns or None: uri for ns, uri in nsmap.items()} if nsmap else None,  # type: ignore[misc]
    )

    return typing.cast(std_etree.Element, element)  # lxml Element copies xml.etree.Element interface



import abc
import dataclasses as dc
import datetime as dt
import ipaddress
from copy import deepcopy
from decimal import Decimal
from enum import Enum, IntEnum
from inspect import isclass
from typing import Any, Dict, List, Mapping, Optional, Sized, Type

import pydantic as pd

import pydantic_xml as pxml
from pydantic_xml import errors

from .backend import create_element, etree
from .utils import NsMap, QName, merge_nsmaps


class XmlEncoder:
    """
    Xml data encoder.
    """

    def encode(self, obj: Any) -> str:
        """
        Encodes provided object into a string

        :param obj: object to be encoded
        :return: encoded object
        """

        if isinstance(obj, str):
            return obj
        if isinstance(obj, (int, float, Decimal)):
            return str(obj)
        if isinstance(obj, bool):
            return str(obj).lower()
        if isinstance(obj, (dt.datetime, dt.date, dt.time)):
            return obj.isoformat()
        if isinstance(
            obj, (
                ipaddress.IPv4Address,
                ipaddress.IPv6Address,
                ipaddress.IPv4Network,
                ipaddress.IPv6Network,
                ipaddress.IPv4Interface,
                ipaddress.IPv6Interface,
            ),
        ):
            return str(obj)
        if isinstance(obj, Enum):
            return self.encode(obj.value)

        return self.default(obj)

    def default(self, obj: Any) -> str:
        raise TypeError(f'Object of type {obj.__class__.__name__} is not XML serializable')


DEFAULT_ENCODER = XmlEncoder()


class Location(IntEnum):
    """
    Field data location.
    """

    MISSING = 0  # field location is not provided
    ELEMENT = 1  # field data is located at xml element
    ATTRIBUTE = 2  # field data is located at xml attribute
    WRAPPED = 3  # field data is wrapped by an element


class PydanticShapeType(IntEnum):
    """
    Pydantic shape type.
    """

    UNKNOWN = 0
    SCALAR = 1
    HOMOGENEOUS = 2
    HETEROGENEOUS = 3
    MAPPING = 4

    __SHAPE_TYPES__: Dict[int, int] = {
        pd.fields.SHAPE_SINGLETON:      SCALAR,

        pd.fields.SHAPE_LIST:           HOMOGENEOUS,
        pd.fields.SHAPE_SET:            HOMOGENEOUS,
        pd.fields.SHAPE_TUPLE_ELLIPSIS: HOMOGENEOUS,
        pd.fields.SHAPE_SEQUENCE:       HOMOGENEOUS,
        pd.fields.SHAPE_ITERABLE:       HOMOGENEOUS,
        pd.fields.SHAPE_FROZENSET:      HOMOGENEOUS,
        pd.fields.SHAPE_DEQUE:          HOMOGENEOUS,

        pd.fields.SHAPE_TUPLE:          HETEROGENEOUS,

        pd.fields.SHAPE_MAPPING:        MAPPING,
        pd.fields.SHAPE_DICT:           MAPPING,
        pd.fields.SHAPE_DEFAULTDICT:    MAPPING,

        pd.fields.SHAPE_GENERIC:        UNKNOWN,
    }

    @classmethod
    def from_shape(cls, shape: int) -> 'PydanticShapeType':
        return cls(cls.__SHAPE_TYPES__.get(shape, cls.UNKNOWN))


def find_element_or_create(root: etree.Element, name: str) -> etree.Element:
    if (sub_element := root.find(name)) is None:
        sub_element = etree.SubElement(root, name)

    return sub_element


class Serializer(abc.ABC):
    """
    Base field serializer/deserializer.
    """

    @dc.dataclass(frozen=True)
    class Context:
        parent_is_root: bool = False
        parent_ns: Optional[str] = None
        parent_nsmap: Optional[NsMap] = None

        entity_name: Optional[str] = None
        entity_ns: Optional[str] = None
        entity_nsmap: Optional[NsMap] = None

    @abc.abstractmethod
    def serialize(self, element: etree.Element, value: Any, *, encoder: XmlEncoder, skip_empty: bool = False) -> Any:
        """
        Serializes a value into an xml element.

        :param element: element serialized value should be added to
        :param value: value to be serialized
        :param encoder: xml encoder to be used to serialize the value
        :param skip_empty: skip empty elements (elements without sub-elements, attributes and text, Nones)
        """

    @abc.abstractmethod
    def deserialize(self, root: etree.Element) -> Any:
        """
        Deserializes a value from an xml element.

        :param root: element deserialized value should be fetched to
        :return: deserialized value
        """

    @classmethod
    def build_field_serializer(
            cls,
            model: Type['pxml.BaseXmlModel'],
            model_field: pd.fields.ModelField,
            ctx: Context,
    ) -> 'Serializer':
        field_type = model_field.type_
        field_info = model_field.field_info

        shape_type = PydanticShapeType.from_shape(model_field.shape)
        if shape_type is PydanticShapeType.UNKNOWN:
            raise TypeError(f"fields of type {model_field.type_} are not supported")

        if isclass(field_type) and issubclass(field_type, pxml.BaseXmlModel):
            is_model_field = True
        else:
            is_model_field = False

        if isinstance(field_info, pxml.XmlElementInfo):
            field_location = Location.ELEMENT
            ctx = dc.replace(ctx, entity_name=field_info.tag, entity_ns=field_info.ns, entity_nsmap=field_info.nsmap)
        elif isinstance(field_info, pxml.XmlAttributeInfo):
            field_location = Location.ATTRIBUTE
            ctx = dc.replace(ctx, entity_name=field_info.name, entity_ns=field_info.ns)
        elif isinstance(field_info, pxml.XmlWrapperInfo):
            field_location = Location.WRAPPED
            ctx = dc.replace(ctx, entity_name=field_info.path, entity_ns=field_info.ns, entity_nsmap=field_info.nsmap)
        else:
            field_location = Location.MISSING
            ctx = dc.replace(
                ctx,
                parent_ns=ctx.entity_ns or ctx.parent_ns,
                parent_nsmap=merge_nsmaps(ctx.entity_nsmap, ctx.parent_nsmap),
                entity_name=None,
                entity_ns=None,
                entity_nsmap=None,
            )

        if field_location is Location.WRAPPED:
            return WrappedSerializerFactory.build(model, model_field, ctx)
        elif shape_type is PydanticShapeType.SCALAR and not is_model_field:
            return PrimitiveTypeSerializerFactory.build(model, model_field, field_location, ctx)
        elif shape_type is PydanticShapeType.SCALAR and is_model_field:
            return ModelSerializerFactory.build(model, model_field, field_location, ctx)
        elif shape_type is PydanticShapeType.MAPPING:
            return MappingSerializerFactory.build(model, model_field, field_location, ctx)
        elif shape_type is PydanticShapeType.HOMOGENEOUS:
            return HomogeneousSerializerFactory.build(model, model_field, field_location, ctx)
        elif shape_type is PydanticShapeType.HETEROGENEOUS:
            return HeterogeneousSerializerFactory.build(model, model_field, field_location, ctx)
        else:
            raise AssertionError("unreachable")


class PrimitiveTypeSerializerFactory:
    """
    Primitive type serializer factory.
    """

    class TextSerializer(Serializer):
        def serialize(
                self, element: etree.Element, value: Any, *, encoder: XmlEncoder, skip_empty: bool = False,
        ) -> Optional[etree.Element]:
            if value is None and skip_empty:
                return element

            encoded = encoder.encode(value)
            element.text = encoded
            return element

        def deserialize(self, element: etree.Element) -> Optional[str]:
            return element.text or None

    class AttributeSerializer(Serializer):
        def __init__(
                self, model: Type['pxml.BaseXmlModel'], model_field: pd.fields.ModelField, ctx: Serializer.Context,
        ):
            ns_attrs = model.__xml_ns_attrs__
            name = ctx.entity_name or model_field.alias
            ns = ctx.entity_ns or (ctx.parent_ns if ns_attrs else None)
            nsmap = ctx.parent_nsmap

            self.attr_name = QName.from_alias(tag=name, ns=ns, nsmap=nsmap, is_attr=True).uri

        def serialize(
                self, element: etree.Element, value: Any, *, encoder: XmlEncoder, skip_empty: bool = False,
        ) -> Optional[etree.Element]:
            if value is None and skip_empty:
                return element

            encoded = encoder.encode(value)
            element.set(self.attr_name, encoded)

            return element

        def deserialize(self, element: etree.Element) -> Optional[str]:
            return element.get(self.attr_name)

    class ElementSerializer(Serializer):
        def __init__(self, model_field: pd.fields.ModelField, ctx: Serializer.Context):
            name = ctx.entity_name or model_field.alias
            ns = ctx.entity_ns or ctx.parent_ns
            nsmap = merge_nsmaps(ctx.entity_nsmap, ctx.parent_nsmap)
            self.element_name = QName.from_alias(tag=name, ns=ns, nsmap=nsmap).uri

        def serialize(
                self, element: etree.Element, value: Any, *, encoder: XmlEncoder, skip_empty: bool = False,
        ) -> Optional[etree.Element]:
            if value is None and skip_empty:
                return element

            encoded = encoder.encode(value)

            sub_element = find_element_or_create(element, self.element_name)
            sub_element.text = encoded
            return sub_element

        def deserialize(self, element: etree.Element) -> Any:
            return element.findtext(self.element_name)

    @classmethod
    def build(
            cls,
            model: Type['pxml.BaseXmlModel'],
            model_field: pd.fields.ModelField,
            field_location: Location,
            ctx: Serializer.Context,
    ) -> 'Serializer':
        if field_location is Location.ELEMENT:
            return cls.ElementSerializer(model_field, ctx)
        elif field_location is Location.ATTRIBUTE:
            return cls.AttributeSerializer(model, model_field, ctx)
        elif field_location is Location.MISSING:
            return cls.TextSerializer()
        else:
            raise AssertionError("unreachable")


class ModelSerializerFactory:
    """
    Model serializer factory.
    """

    class RootSerializer(Serializer):
        def __init__(
                self,
                model: Type['pxml.BaseXmlModel'],
                ctx: Serializer.Context,
        ):
            name = ctx.entity_name or model.__xml_tag__ or model.__name__
            ns = ctx.entity_ns or model.__xml_ns__
            nsmap = merge_nsmaps(ctx.entity_nsmap, model.__xml_nsmap__, ctx.parent_nsmap)
            is_root = model.__custom_root_type__
            ctx = dc.replace(
                ctx,
                parent_ns=ns,
                parent_nsmap=nsmap,
                parent_is_root=is_root,
            )

            self.nsmap = nsmap
            self.is_root = is_root
            self.element_name = QName.from_alias(tag=name, ns=ns, nsmap=nsmap).uri
            self.field_serializers = {
                model_subfield.alias: self.build_field_serializer(model, model_subfield, ctx)
                for field_name, model_subfield in model.__fields__.items()
            }

        def serialize(
                self, element: Optional[etree.Element], value: Any, *, encoder: XmlEncoder, skip_empty: bool = False,
        ) -> Optional[etree.Element]:
            if value is None:
                return None

            if element is None:
                element = create_element(self.element_name, nsmap=self.nsmap)

            for field_name, field_serializer in self.field_serializers.items():
                field_serializer.serialize(element, getattr(value, field_name), encoder=encoder, skip_empty=skip_empty)

            return element

        def deserialize(self, element: etree.Element) -> Any:
            result = {
                field_name: field_value
                for field_name, field_serializer in self.field_serializers.items()
                if (field_value := field_serializer.deserialize(element)) is not None
            }
            if self.is_root:
                return result['__root__']
            else:
                return result

    class ElementSerializer(Serializer):

        def __init__(
                self,
                model_field: Optional[pd.fields.ModelField],
                model: Type['pxml.BaseXmlModel'],
                ctx: Serializer.Context,
        ):
            field_name = model_field.alias if model_field else None
            name = ctx.entity_name or model.__xml_tag__ or field_name or model.__name__
            ns = ctx.entity_ns or model.__xml_ns__
            nsmap = merge_nsmaps(ctx.entity_nsmap, model.__xml_nsmap__, ctx.parent_nsmap)

            self.nsmap = nsmap
            self.element_name = QName.from_alias(tag=name, ns=ns, nsmap=nsmap).uri
            self.model = model

        def serialize(
                self, element: etree.Element, value: Any, *, encoder: XmlEncoder, skip_empty: bool = False,
        ) -> Optional[etree.Element]:
            assert self.model.__xml_serializer__ is not None, "model is partially initialized"

            if value is None:
                return None

            sub_element = create_element(self.element_name, nsmap=self.nsmap)

            self.model.__xml_serializer__.serialize(sub_element, value, encoder=encoder, skip_empty=skip_empty)

            if not skip_empty or sub_element.text or sub_element.attrib or len(sub_element) != 0:
                element.append(sub_element)
                return sub_element
            else:
                return None

        def deserialize(self, element: etree.Element) -> Optional[Dict[str, Any]]:
            assert self.model.__xml_serializer__ is not None, "model is partially initialized"

            if (sub_element := element.find(self.element_name)) is not None:
                return self.model.__xml_serializer__.deserialize(sub_element)
            else:
                return None

    class DeferredSerializer(Serializer):

        def __init__(self, model: Type['pxml.BaseXmlModel']):
            self.model = model

        def serialize(
                self, element: etree.Element, value: Any, *, encoder: XmlEncoder, skip_empty: bool = False,
        ) -> Optional[etree.Element]:
            assert self.model.__xml_serializer__ is not None, "model is partially initialized"

            return self.model.__xml_serializer__.serialize(element, value, encoder=encoder, skip_empty=skip_empty)

        def deserialize(self, element: etree.Element) -> Optional[Dict[str, Any]]:
            assert self.model.__xml_serializer__ is not None, "model is partially initialized"

            return self.model.__xml_serializer__.deserialize(element)

    @classmethod
    def from_model(cls, model: Type['pxml.BaseXmlModel']) -> 'RootSerializer':
        return cls.RootSerializer(model, Serializer.Context(parent_is_root=True))

    @classmethod
    def build(
            cls,
            model: Type['pxml.BaseXmlModel'],
            model_field: pd.fields.ModelField,
            field_location: Location,
            ctx: Serializer.Context,
    ) -> 'Serializer':
        sub_model = model_field.type_
        ctx = dc.replace(
            ctx,
            parent_ns=ctx.parent_ns or model.__xml_ns__,
            parent_nsmap=merge_nsmaps(ctx.parent_nsmap, model.__xml_nsmap__),
        )

        if field_location is Location.ELEMENT:
            return cls.ElementSerializer(model_field, sub_model, ctx)
        elif not ctx.parent_is_root and field_location is Location.MISSING:
            return cls.ElementSerializer(model_field, sub_model, ctx)
        elif ctx.parent_is_root and field_location is Location.MISSING:
            return cls.DeferredSerializer(sub_model)
        elif field_location is Location.ATTRIBUTE:
            raise errors.ModelFieldError(
                model.__name__, model_field.name, "attributes of model type are not supported",
            )
        else:
            raise AssertionError("unreachable")


class MappingSerializerFactory:
    """
    Mapping type serializer factory.
    """

    class BaseSerializer(Serializer, abc.ABC):
        def __init__(
                self, model: Type['pxml.BaseXmlModel'], model_field: pd.fields.ModelField, ctx: Serializer.Context,
        ):
            name = ctx.entity_name or model_field.alias

            self.parent_ns = ns = ctx.entity_ns or ctx.parent_ns
            self.parent_nsmap = nsmap = merge_nsmaps(ctx.entity_nsmap, ctx.parent_nsmap)
            self.element_name = QName.from_alias(tag=name, ns=ns, nsmap=nsmap).uri
            self.ns_attrs = model.__xml_ns_attrs__

    class AttributesSerializer(BaseSerializer):
        def serialize(
                self, element: etree.Element, value: Mapping[str, Any], *, encoder: XmlEncoder, skip_empty: bool = False
        ) -> Optional[etree.Element]:
            if value is None:
                return element

            if skip_empty and len(value) == 0:
                return element

            if self.ns_attrs:
                ns = self.parent_nsmap.get(self.parent_ns) if self.parent_ns else None
                element.attrib.update({
                    QName(tag=attr, ns=ns).uri: encoder.encode(val)
                    for attr, val in value.items()
                })
            else:
                element.attrib.update({
                    attr: encoder.encode(val)
                    for attr, val in value.items()
                })

            return element

        def deserialize(self, element: etree.Element) -> Dict[str, str]:
            if self.ns_attrs:
                return {QName.from_uri(attr).tag: val for attr, val in element.attrib.items()}
            else:
                return {attr: val for attr, val in element.attrib.items()}

    class ElementSerializer(BaseSerializer):
        def serialize(
                self, element: etree.Element, value: Mapping[str, Any], *, encoder: XmlEncoder, skip_empty: bool = False
        ) -> etree.Element:
            if skip_empty and len(value) == 0:
                return element

            sub_element = find_element_or_create(element, self.element_name)
            if self.ns_attrs:
                ns = self.parent_nsmap.get(self.parent_ns) if self.parent_ns else None
                sub_element.attrib.update({
                    QName(tag=attr, ns=ns).uri: encoder.encode(val)
                    for attr, val in value.items()
                })
            else:
                sub_element.attrib.update({
                    attr: encoder.encode(val)
                    for attr, val in value.items()
                })

            return sub_element

        def deserialize(self, element: etree.Element) -> Optional[Dict[str, str]]:
            if (sub_element := element.find(self.element_name)) is not None:
                if self.ns_attrs:
                    return {
                        QName.from_uri(attr).tag: val
                        for attr, val in sub_element.attrib.items()
                    }
                else:
                    return {
                        attr: val
                        for attr, val in sub_element.attrib.items()
                    }
            else:
                return None

    @classmethod
    def build(
            cls,
            model: Type['pxml.BaseXmlModel'],
            model_field: pd.fields.ModelField,
            field_location: Location,
            ctx: Serializer.Context,
    ) -> 'Serializer':
        assert model_field.sub_fields is not None, "unexpected model field"
        value_field = model_field.sub_fields[0]
        value_type = value_field.type_
        if PydanticShapeType.from_shape(value_field.shape) is not PydanticShapeType.SCALAR:
            raise errors.ModelFieldError(
                model.__name__, model_field.name, "mapping value should be of scalar type",
            )

        if isclass(value_type) and issubclass(value_type, pxml.BaseXmlModel):
            raise errors.ModelFieldError(
                model.__name__, model_field.name, "mapping value types can't be models",
            )

        if field_location is Location.ELEMENT:
            return cls.ElementSerializer(model, model_field, ctx)
        elif field_location is Location.MISSING:
            return cls.AttributesSerializer(model, model_field, ctx)
        elif field_location is Location.ATTRIBUTE:
            raise errors.ModelFieldError(
                model.__name__, model_field.name, "attributes of mapping type are not supported",
            )
        else:
            raise AssertionError("unreachable")


class HomogeneousSerializerFactory:
    """
    Homogeneous collection type serializer factory.
    """

    class ElementSerializer(Serializer):
        def __init__(
                self, model: Type['pxml.BaseXmlModel'], model_field: pd.fields.ModelField, ctx: Serializer.Context,
        ):
            assert model_field.sub_fields is not None, "unexpected model field"

            name = ctx.entity_name or model_field.alias
            ns = ctx.entity_ns or ctx.parent_ns
            nsmap = merge_nsmaps(ctx.entity_nsmap, ctx.parent_nsmap)

            self.element_name = QName.from_alias(tag=name, ns=ns, nsmap=nsmap).uri

            item_field = deepcopy(model_field.sub_fields[0])
            item_field.name = model_field.alias
            self.serializer = self.build_field_serializer(
                model,
                item_field,
                dc.replace(
                    ctx,
                    parent_is_root=True,
                    parent_ns=ns,
                    parent_nsmap=nsmap,
                ),
            )

        def serialize(
                self, element: etree.Element, value: List[Any], *, encoder: XmlEncoder, skip_empty: bool = False,
        ) -> Optional[etree.Element]:
            if value is None:
                return element

            if skip_empty and len(value) == 0:
                return element

            for val in value:
                sub_element = etree.SubElement(element, self.element_name)
                self.serializer.serialize(sub_element, val, encoder=encoder, skip_empty=skip_empty)

            return element

        def deserialize(self, element: etree.Element) -> Optional[List[Any]]:
            return [
                self.serializer.deserialize(sub_element)
                for sub_element in element.findall(self.element_name)
            ]

    @classmethod
    def build(
            cls,
            model: Type['pxml.BaseXmlModel'],
            model_field: pd.fields.ModelField,
            field_location: Location,
            ctx: Serializer.Context,
    ) -> 'Serializer':
        is_root = model.__custom_root_type__

        assert model_field.sub_fields is not None, "unexpected model field"
        item_field = model_field.sub_fields[0]

        if PydanticShapeType.from_shape(item_field.shape) in (
            PydanticShapeType.HOMOGENEOUS,
            PydanticShapeType.HETEROGENEOUS,
        ):
            raise errors.ModelFieldError(
                model.__name__, model_field.name, "collection elements can't be of collection type",
            )

        if is_root and field_location is Location.MISSING:
            raise errors.ModelFieldError(
                model.__name__, model_field.name, "root model collections should be marked as elements",
            )

        if field_location is Location.ELEMENT:
            return cls.ElementSerializer(model, model_field, ctx)
        elif field_location is Location.MISSING:
            return cls.ElementSerializer(model, model_field, ctx)
        elif field_location is Location.ATTRIBUTE:
            raise errors.ModelFieldError(
                model.__name__, model_field.name, "attributes of collection type are not supported",
            )
        else:
            raise AssertionError("unreachable")


class HeterogeneousSerializerFactory:
    """
    Heterogeneous collection type serializer factory.
    """

    class ElementSerializer(Serializer):
        def __init__(
                self, model: Type['pxml.BaseXmlModel'], model_field: pd.fields.ModelField, ctx: Serializer.Context,
        ):
            assert model_field.sub_fields is not None, "unexpected model field"

            name = ctx.entity_name or model_field.alias
            ns = ctx.entity_ns or ctx.parent_ns
            nsmap = merge_nsmaps(ctx.entity_nsmap, ctx.parent_nsmap)

            self.element_name = QName.from_alias(tag=name, ns=ns, nsmap=nsmap).uri

            self.serializers = []
            for sub_field in model_field.sub_fields:
                sub_field = deepcopy(sub_field)
                sub_field.name = model_field.alias

                self.serializers.append(
                    self.build_field_serializer(
                        model,
                        sub_field,
                        dc.replace(
                            ctx,
                            parent_is_root=True,
                            parent_ns=ns,
                            parent_nsmap=nsmap,
                        ),
                    ),
                )

        def serialize(
                self, element: etree.Element, value: List[Any], *, encoder: XmlEncoder, skip_empty: bool = False,
        ) -> Optional[etree.Element]:
            if value is None:
                return element

            if skip_empty and len(value) == 0:
                return element

            for serializer, val in zip(self.serializers, value):
                sub_element = etree.SubElement(element, self.element_name)
                serializer.serialize(sub_element, val, encoder=encoder, skip_empty=skip_empty)

            return element

        def deserialize(self, element: etree.Element) -> Optional[List[Any]]:
            return [
                serializer.deserialize(sub_element)
                for serializer, sub_element in zip(self.serializers, element.findall(self.element_name))
            ]

    @classmethod
    def build(
            cls,
            model: Type['pxml.BaseXmlModel'],
            model_field: pd.fields.ModelField,
            field_location: Location,
            ctx: Serializer.Context,
    ) -> 'Serializer':
        is_root = model.__custom_root_type__

        assert model_field.sub_fields is not None, "unexpected model field"
        for item_field in model_field.sub_fields:
            if PydanticShapeType.from_shape(item_field.shape) in (
                PydanticShapeType.HOMOGENEOUS,
                PydanticShapeType.HETEROGENEOUS,
            ):
                raise errors.ModelFieldError(
                    model.__name__, model_field.name, "collection elements can't be of collection type",
                )

            if is_root and field_location is Location.MISSING:
                raise errors.ModelFieldError(
                    model.__name__, model_field.name, "root model collections should be marked as elements",
                )

        if field_location is Location.ELEMENT:
            return cls.ElementSerializer(model, model_field, ctx)
        elif field_location is Location.MISSING:
            return cls.ElementSerializer(model, model_field, ctx)
        elif field_location is Location.ATTRIBUTE:
            raise errors.ModelFieldError(
                model.__name__, model_field.name, "attributes of collection type are not supported",
            )
        else:
            raise AssertionError("unreachable")


class WrappedSerializerFactory:
    """
    Wrapped serializer factory.
    """

    class ElementPathSerializer(Serializer):
        def __init__(
                self, model: Type['pxml.BaseXmlModel'], model_field: pd.fields.ModelField, ctx: Serializer.Context,
        ):
            path = ctx.entity_name
            ns = ctx.entity_ns or ctx.parent_ns
            nsmap = merge_nsmaps(ctx.entity_nsmap, ctx.parent_nsmap)

            model_field = deepcopy(model_field)
            field_info = model_field.field_info

            assert path is not None, "path is not provided"
            assert isinstance(field_info, pxml.XmlWrapperInfo)
            model_field.field_info = field_info.entity or pd.fields.FieldInfo()

            self.path = tuple(QName.from_alias(tag=part, ns=ns, nsmap=nsmap).uri for part in path.split('/'))
            self.serializer = self.build_field_serializer(
                model,
                model_field,
                ctx=dc.replace(
                    ctx,
                    parent_is_root=True,
                    parent_ns=ns,
                    parent_nsmap=nsmap,
                ),
            )

        def serialize(
                self, element: etree.Element, value: Any, *, encoder: XmlEncoder, skip_empty: bool = False,
        ) -> Optional[etree.Element]:
            if value is None:
                return element

            if skip_empty and isinstance(value, Sized) and len(value) == 0:
                return element

            for part in self.path:
                element = find_element_or_create(element, part)

            self.serializer.serialize(element, value, encoder=encoder, skip_empty=skip_empty)

            return element

        def deserialize(self, element: etree.Element) -> Optional[Any]:
            if (sub_element := element.find('/'.join(self.path))) is not None:
                return self.serializer.deserialize(sub_element)
            else:
                return None

    @classmethod
    def build(
            cls,
            model: Type['pxml.BaseXmlModel'],
            model_field: pd.fields.ModelField,
            ctx: Serializer.Context,
    ) -> 'Serializer':
        return cls.ElementPathSerializer(model, model_field, ctx)





class BaseError(Exception):
    """
    Base package error.
    """


class ModelError(BaseError):
    """
    Model definition error.
    """


class ModelFieldError(BaseError):
    """
    Model field definition error.
    """

    def __init__(self, model_name: str, field_name: str, message: str):
        self.model_name = model_name
        self.field_name = field_name
        self.message = message

        super().__init__(f"{model_name}.{field_name} field type incorrect: {message}")



from pydantic_xml import config

if config.FORCE_STD_XML:
    from .std import *  # noqa: F403
else:
    try:
        from .lxml import *  # type: ignore[no-redef]  # noqa: F403
    except ImportError:
        from .std import *  # noqa: F403



import os


def strtobool(val: str) -> bool:
    """
    Converts a string representation of boolean type to true or false.
    """

    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    else:
        raise ValueError


REGISTER_NS_PREFIXES = strtobool(os.environ.get('REGISTER_NS_PREFIXES', 'true'))
FORCE_STD_XML = strtobool(os.environ.get('FORCE_STD_XML', 'false'))



import dataclasses as dc
import re
from collections import ChainMap
from typing import Dict, Optional, cast

from .backend import etree

NsMap = Dict[str, str]


@dc.dataclass(frozen=True)
class QName:
    """
    XML entity qualified name.

    :param tag: entity tag
    :param ns: entity namespace
    """

    tag: str
    ns: Optional[str]

    @classmethod
    def from_uri(cls, uri: str) -> 'QName':
        """
        Creates `QName` from uri.

        :param uri: entity uri in format '{namespace}tag'
        :return: qualified name
        """

        if (m := re.match(r'({(.*)})?(.*)', uri)) is None:
            raise ValueError

        return cls(tag=m[3], ns=m[2])

    @classmethod
    def from_alias(
            cls, tag: str, ns: Optional[str] = None, nsmap: Optional[NsMap] = None, is_attr: bool = False,
    ) -> 'QName':
        """
        Creates `QName` from namespace alias.

        :param tag: entity tag
        :param ns: xml namespace alias
        :param nsmap: xml namespace mapping
        :param is_attr: is the tag of attribute type
        :return: qualified name
        """

        if not is_attr or ns is not None:
            ns = nsmap.get(ns or '') if nsmap else None

        return QName(tag=tag, ns=ns)

    @property
    def uri(self) -> str:
        return '{%s}%s' % (self.ns, self.tag) if self.ns else self.tag

    def __str__(self) -> str:
        return self.uri


def merge_nsmaps(*maps: Optional[NsMap]) -> NsMap:
    """
    Merges multiple namespace maps into s single one respecting provided order.

    :param maps: namespace maps
    :return: merged namespace
    """

    return cast(NsMap, ChainMap(*(nsmap for nsmap in maps if nsmap)))


def register_nsmap(nsmap: NsMap) -> None:
    """
    Registers namespaces prefixes from the map.
    """

    for prefix, uri in nsmap.items():
        if prefix != '':  # skip default namespace
            etree.register_namespace(prefix, uri)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydantic_xml', 'pydantic_xml.backend']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.0,<2.0.0']

extras_require = \
{'lxml': ['lxml>=4.9.1,<5.0.0']}

setup_kwargs = {
    'name': 'pydantic-xml',
    'version': '0.5.0',
    'description': 'pydantic xml serialization/deserialization extension',
    'long_description': '[![Downloads][download-badge]][download-url]\n[![License][licence-badge]][licence-url]\n[![Python Versions][python-version-badge]][python-version-url]\n[![Test Status][test-badge]][test-url]\n[![Test Coverage][test-cov-badge]][test-cov-url]\n\n[download-badge]: https://static.pepy.tech/personalized-badge/pydantic-xml?period=month&units=international_system&left_color=grey&right_color=orange&left_text=Downloads/month\n[download-url]: https://pepy.tech/project/pydantic-xml\n[licence-badge]: https://img.shields.io/badge/license-Unlicense-blue.svg\n[licence-url]: https://github.com/dapper91/pydantic-xml/blob/master/LICENSE\n[python-version-badge]: https://img.shields.io/pypi/pyversions/pydantic-xml.svg\n[python-version-url]: https://pypi.org/project/pydantic-xml\n[test-badge]: https://github.com/dapper91/pydantic-xml/actions/workflows/test.yml/badge.svg?branch=master\n[test-url]: https://github.com/dapper91/pydantic-xml/actions/workflows/test.yml\n[test-cov-badge]: https://codecov.io/gh/dapper91/pydantic-xml/branch/master/graph/badge.svg\n[test-cov-url]: https://codecov.io/gh/dapper91/pydantic-xml\n\n# pydantic-xml\n\npydantic xml extension\n\n\n## Installation\n\n```commandline\npip install pydantic-xml[lxml]\n```\n\n* **lxml** - use `lxml` instead of standard `xml.etree.ElementTree` library\n\n## Quickstart\n\n`pydantic-xml` is a pydantic extension implementing model xml serialization/deserialization.\nIt is closely integrated with `pydantic` which means it supports most of the `pydantic` features.\n\nAll xml serializable/deserializable models should be inherited from `BaseXmlModel` base class.\n`BaseXmlModel` implements method `to_xml` to serialize an object to an xml string and `from_xml` to deserialize it.\nModel field data could be extracted from xml attribute, element or text. Field data location is derived\nusing the following rules:\n\n* fields of primitive types (`int`, `float`, `str`, `datetime`, ...) are extracted from element text by default.\n  To alter the default behaviour the field should be marked as `attr` or `element`.\n\n```python\nclass Company(BaseXmlModel):\n    trade_name: str = attr(name=\'trade-name\') # extracted from the \'trade-name\' attribute\n    website: HttpUrl = element() # extracted from the \'website\' element text\n    description: str # extracted from the root element text\n```\n\n```xml\n<Company trade-name="SpaceX">\n    company description text\n    <website>https://www.spacex.com</website>\n</Company>\n```\n\n* fields of model types are extracted from an element.\n\n```python\nclass Headquarters(BaseXmlModel):\n    country: str = element()\n    state: str = element()\n    city: str = element()\n\n\nclass Company(BaseXmlModel):\n    headquarters: Headquarters\n```\n\n```xml\n<Company>\n    <headquarters>\n        <country>US</country>\n        <state>California</state>\n        <city>Hawthorne</city>\n    </headquarters>\n</Company>\n```\n\n* fields of mapping types are extracted from the current element attributes (by default)\n  or from a nested element attributes if the field is marked as `element`.\n\n```python\nclass Company(BaseXmlModel):\n    company_attributes: Dict[str, str]\n    founder_attributes: Dict[str, str] = element(tag=\'Founder\')\n```\n\n```xml\n<Company trade-name="SpaceX" type="Private">\n    <Founder name="Elon" surname="Musk"/>\n</Company>\n```\n\n* fields of collection types (`list`, `tuple`, `set`, ...) are extracted from multiple elements with the same name.\n\n```python\nclass Product(BaseXmlModel):\n    status: Literal[\'running\', \'development\'] = attr()\n    launched: Optional[int] = attr()\n    title: str\n\n\nclass Company(BaseXmlModel):\n    products: List[Product] = element(tag=\'product\')\n```\n\n```xml\n<Company>\n    <product status="running" launched="2013">Several launch vehicles</product>\n    <product status="running" launched="2019">Starlink</product>\n    <product status="development">Starship</product>\n</Company>\n```\n\n* wrapped fields are extracted from a sub-element located at the provided path.\n\n```python\nclass Social(BaseXmlModel):\n    type: str = attr()\n    url: str\n\n\nclass Company(BaseXmlModel):\n    socials: List[Social] = wrapped(\'contacts/socials\', element(tag=\'social\'))\n```\n\n```xml\n<Company>\n    <contacts>\n        <socials>\n            <social type="linkedin">https://www.linkedin.com/company/spacex</social>\n            <social type="twitter">https://twitter.com/spacex</social>\n            <social type="youtube">https://www.youtube.com/spacex</social>\n        </socials>\n    </contacts>\n</Company>\n```\n\nThe following example illustrates all previously described rules in conjunction with some `pydantic` feature:\n\n*company.xml:*\n\n```xml\n<Company trade-name="SpaceX" type="Private" xmlns:pd="http://www.test.com/prod">\n    <Founder name="Elon" surname="Musk"/>\n    <Founded>2002-03-14</Founded>\n    <Employees>12000</Employees>\n    <WebSite>https://www.spacex.com</WebSite>\n\n    <Industries>\n        <Industry>space</Industry>\n        <Industry>communications</Industry>\n    </Industries>\n\n    <key-people>\n        <person position="CEO" name="Elon Musk"/>\n        <person position="CTO" name="Elon Musk"/>\n        <person position="COO" name="Gwynne Shotwell"/>\n    </key-people>\n\n    <hq:headquarters xmlns:hq="http://www.test.com/hq">\n        <hq:country>US</hq:country>\n        <hq:state>California</hq:state>\n        <hq:city>Hawthorne</hq:city>\n    </hq:headquarters>\n\n    <co:contacts xmlns:co="http://www.test.com/contact" >\n        <co:socials>\n            <co:social co:type="linkedin">https://www.linkedin.com/company/spacex</co:social>\n            <co:social co:type="twitter">https://twitter.com/spacex</co:social>\n            <co:social co:type="youtube">https://www.youtube.com/spacex</co:social>\n        </co:socials>\n    </co:contacts>\n\n    <pd:product pd:status="running" pd:launched="2013">Several launch vehicles</pd:product>\n    <pd:product pd:status="running" pd:launched="2019">Starlink</pd:product>\n    <pd:product pd:status="development">Starship</pd:product>\n</Company>\n```\n\n*main.py:*\n\n```python\nfrom datetime import date\nfrom enum import Enum\nfrom typing import Dict, List, Optional, Set, Literal, Tuple\n\nimport pydantic as pd\nfrom pydantic import conint, HttpUrl\n\nfrom pydantic_xml import BaseXmlModel, attr, element, wrapped\n\nNSMAP = {\n    \'co\': \'http://www.test.com/contact\',\n    \'hq\': \'http://www.test.com/hq\',\n    \'pd\': \'http://www.test.com/prod\',\n}\n\n\nclass Headquarters(BaseXmlModel, ns=\'hq\', nsmap=NSMAP):\n    country: str = element()\n    state: str = element()\n    city: str = element()\n\n    @pd.validator(\'country\')\n    def validate_country(cls, value: str) -> str:\n        if len(value) > 2:\n            raise ValueError(\'country must be of 2 characters\')\n        return value\n\n\nclass Industries(BaseXmlModel):\n    __root__: Set[str] = element(tag=\'Industry\')\n\n\nclass Social(BaseXmlModel, ns_attrs=True, ns=\'co\', nsmap=NSMAP):\n    type: str = attr()\n    url: str\n\n\nclass Product(BaseXmlModel, ns_attrs=True, ns=\'pd\', nsmap=NSMAP):\n    status: Literal[\'running\', \'development\'] = attr()\n    launched: Optional[int] = attr()\n    title: str\n\n\nclass Person(BaseXmlModel):\n    name: str = attr()\n\n\nclass CEO(Person):\n    position: Literal[\'CEO\'] = attr()\n\n\nclass CTO(Person):\n    position: Literal[\'CTO\'] = attr()\n\n\nclass COO(Person):\n    position: Literal[\'COO\'] = attr()\n\n\nclass Company(BaseXmlModel, tag=\'Company\', nsmap=NSMAP):\n    class CompanyType(str, Enum):\n        PRIVATE = \'Private\'\n        PUBLIC = \'Public\'\n\n    trade_name: str = attr(name=\'trade-name\')\n    type: CompanyType = attr()\n    employees: conint(gt=0) = element(tag=\'Employees\')\n    website: HttpUrl = element(tag=\'WebSite\')\n\n    founder: Dict[str, str] = element(tag=\'Founder\')\n    founded: Optional[date] = element(tag=\'Founded\')\n    industries: Industries = element(tag=\'Industries\')\n\n    key_people: Tuple[CEO, CTO, COO] = wrapped(\'key-people\', element(tag=\'person\'))\n    headquarters: Headquarters\n    socials: List[Social] = wrapped(\n        \'contacts/socials\',\n        element(tag=\'social\', default_factory=list),\n        ns=\'co\',\n        nsmap=NSMAP,\n    )\n\n    products: Tuple[Product, ...] = element(tag=\'product\', ns=\'pd\')\n\n\nwith open(\'company.xml\') as file:\n    document = file.read()\n\ncompany = Company.from_xml(document)\n\nprint(company)\n\n```\n\n### Name resolution:\n\n**TBD**\n\n\n### Namespace resolution:\n\n#### Default namespace\n\nTo parse xml documents with a declared default namespace add it to `nsmap`  with an empty key.\nLook at the following example:\n\n```xml\n<headquarters xmlns="http://www.test.com/hq">\n    <country>US</country>\n    <state>California</state>\n    <city>Hawthorne</city>\n</headquarters>\n```\n\n```python\nclass Headquarters(BaseXmlModel, tag=\'headquarters\', nsmap={\'\': \'http://www.test.com/hq\'}):\n    country: str = element()\n    state: str = element()\n    city: str = element()\n```\n\n**TBD**\n\n\n### Generics:\n\n`pydantic` library supports [generic-models](https://pydantic-docs.helpmanual.io/usage/models/#generic-models).\n`pydantic-xml` supports it either. To declare one inherit your model from `BaseGenericXmlModel`,\nthe rest is the same as `pydantic` generic model declaration.\n\nThe following example illustrates it:\n\n*request.xml:*\n\n```xml\n<request service-name="api-gateway"\n         request-id="27765d90-f3ef-426f-be9d-8da2b405b4a9"\n         timestamp="2019-06-12T12:21:34.123+00:00">\n    <auth type="basic">\n        <user>gw</user>\n        <password>secret</password>\n    </auth>\n    <payload type="rpc">\n        <method>crete_event</method>\n        <param name="timestamp">1660892066.1952798</param>\n        <param name="user">admin</param>\n        <param name="method">POST</param>\n        <param name="resource">https://api-gateway.test.com/api/v1/users</param>\n    </payload>\n</request>\n```\n\n*main.py:*\n\n```python\nimport datetime as dt\nfrom typing import Generic, Optional, Tuple, TypeVar\n\nfrom pydantic import HttpUrl\n\nimport pydantic_xml as pxml\n\n\nclass BasicAuth(pxml.BaseXmlModel):\n    type: str = pxml.attr()\n    user: str = pxml.element()\n    password: str = pxml.element()\n\n\nAuthType = TypeVar(\'AuthType\')\nPayloadType = TypeVar(\'PayloadType\')\n\n\nclass Request(pxml.BaseGenericXmlModel, Generic[AuthType, PayloadType], tag=\'request\'):\n    service_name: str = pxml.attr(name=\'service-name\')\n    request_id: str = pxml.attr(name=\'request-id\')\n    timestamp: dt.datetime = pxml.attr()\n\n    auth: Optional[AuthType]\n\n    payload: PayloadType\n\n\nParamsType = TypeVar(\'ParamsType\')\nParamType = TypeVar(\'ParamType\')\n\n\nclass Rpc(pxml.BaseGenericXmlModel, Generic[ParamsType]):\n    class Param(pxml.BaseGenericXmlModel, Generic[ParamType]):\n        name: str = pxml.attr()\n        value: ParamType\n\n    method: str = pxml.element()\n    params: ParamsType = pxml.element(tag=\'param\')\n\n\nwith open(\'request.xml\') as file:\n    xml = file.read()\n\nrequest = Request[\n    BasicAuth,\n    Rpc[\n        Tuple[\n            Rpc.Param[float],\n            Rpc.Param[str],\n            Rpc.Param[str],\n            Rpc.Param[HttpUrl]\n        ]\n    ]\n].from_xml(xml)\n\nprint(request.json(indent=4))\n```\n\n*output:*\n\n```json\n{\n    "service_name": "api-gateway",\n    "request_id": "27765d90-f3ef-426f-be9d-8da2b405b4a9",\n    "timestamp": "2019-06-12T12:21:34.123000+00:00",\n    "auth": {\n        "type": "basic",\n        "user": "gw",\n        "password": "secret"\n    },\n    "payload": {\n        "method": "crete_event",\n        "params": [\n            {\n                "name": "timestamp",\n                "value": 1660892066.1952798\n            },\n            {\n                "name": "user",\n                "value": "admin"\n            },\n            {\n                "name": "method",\n                "value": "POST"\n            },\n            {\n                "name": "resource",\n                "value": "https://api-gateway.test.com/api/v1/users"\n            }\n        ]\n    }\n}\n```\n\n\n### Self-referencing models:\n\n`pydantic` library supports [self-referencing models](https://pydantic-docs.helpmanual.io/usage/postponed_annotations/#self-referencing-models).\n`pydantic-xml` supports it either.\n\n*request.xml:*\n\n```xml\n<Directory Name="root" Mode="rwxr-xr-x">\n    <Directory Name="etc" Mode="rwxr-xr-x">\n        <File Name="passwd" Mode="-rw-r--r--"/>\n        <File Name="hosts" Mode="-rw-r--r--"/>\n        <Directory Name="ssh" Mode="rwxr-xr-x"/>\n    </Directory>\n    <Directory Name="bin" Mode="rwxr-xr-x"/>\n    <Directory Name="usr" Mode="rwxr-xr-x">\n        <Directory Name="bin" Mode="rwxr-xr-x"/>\n    </Directory>\n</Directory>\n```\n\n*main.py:*\n\n```python\nfrom typing import List, Optional\n\nimport pydantic_xml as pxml\n\n\nclass File(pxml.BaseXmlModel, tag="File"):\n    name: str = pxml.attr(name=\'Name\')\n    mode: str = pxml.attr(name=\'Mode\')\n\n\nclass Directory(pxml.BaseXmlModel, tag="Directory"):\n    name: str = pxml.attr(name=\'Name\')\n    mode: str = pxml.attr(name=\'Mode\')\n    dirs: Optional[List[\'Directory\']] = pxml.element(tag=\'Directory\')\n    files: Optional[List[File]] = pxml.element(tag=\'File\', default_factory=list)\n\n\nwith open(\'request.xml\') as file:\n    xml = file.read()\n\nroot = Directory.from_xml(xml)\nprint(root.json(indent=4))\n\n```\n\n*output:*\n\n```json\n{\n    "name": "root",\n    "mode": "rwxr-xr-x",\n    "dirs": [\n        {\n            "name": "etc",\n            "mode": "rwxr-xr-x",\n            "dirs": [\n                {\n                    "name": "ssh",\n                    "mode": "rwxr-xr-x",\n                    "dirs": [],\n                    "files": []\n                }\n            ],\n            "files": [\n                {\n                    "name": "passwd",\n                    "mode": "-rw-r--r--"\n                },\n                {\n                    "name": "hosts",\n                    "mode": "-rw-r--r--"\n                }\n            ]\n        },\n        {\n            "name": "bin",\n            "mode": "rwxr-xr-x",\n            "dirs": [],\n            "files": []\n        },\n        {\n            "name": "usr",\n            "mode": "rwxr-xr-x",\n            "dirs": [\n                {\n                    "name": "bin",\n                    "mode": "rwxr-xr-x",\n                    "dirs": [],\n                    "files": []\n                }\n            ],\n            "files": []\n        }\n    ],\n    "files": []\n}\n```\n\n### JSON\n\nSince `pydantic` supports json serialization, `pydantic-xml` could be used as xml-to-json transcoder:\n\n```python\n...\n\nwith open(\'company.xml\') as file:\n    xml = file.read()\n\ncompany = Company.from_xml(xml)\n\nprint(company.json(indent=4))\n```\n\n*output:*\n\n```json\n{\n    "trade_name": "SpaceX",\n    "type": "Private",\n    "employees": 12000,\n    "website": "https://www.spacex.com",\n    "founder": {\n        "name": "Elon",\n        "surname": "Musk"\n    },\n    "founded": "2002-03-14",\n    "industries": [\n        "space",\n        "communications"\n    ],\n    "key_people": [\n        {\n            "name": "Elon Musk",\n            "position": "CEO"\n        },\n        {\n            "name": "Elon Musk",\n            "position": "CTO"\n        },\n        {\n            "name": "Gwynne Shotwell",\n            "position": "COO"\n        }\n    ],\n    "headquarters": {\n        "country": "US",\n        "state": "California",\n        "city": "Hawthorne"\n    },\n    "socials": [\n        {\n            "type": "linkedin",\n            "url": "https://www.linkedin.com/company/spacex"\n        },\n        {\n            "type": "twitter",\n            "url": "https://twitter.com/spacex"\n        },\n        {\n            "type": "youtube",\n            "url": "https://www.youtube.com/spacex"\n        }\n    ],\n    "products": [\n        {\n            "status": "running",\n            "launched": 2013,\n            "title": "Several launch vehicles"\n        },\n        {\n            "status": "running",\n            "launched": 2019,\n            "title": "Starlink"\n        },\n        {\n            "status": "development",\n            "launched": null,\n            "title": "Starship"\n        }\n    ]\n}\n```\n\n\n### XML parser\n\n`pydantic-xml` tries to use the fastest xml parser in your system. It uses `lxml` if it is installed\nin your environment otherwise falls back to the standard library xml parser.\n\n\n### Custom type serialization\n\nOnly several primitive standard type serialization are supported\n(`str`, `int`, `float`, `Decimal`, `bool`, `datetime`, `date`, `time`).\nTo allow custom types serialization `BaseXmlModel.to_xml` method takes an optional encoder as an argument.\nA custom one should be inherited from `XmlEncoder`.\n\nThe following example illustrate how to implement `bytes` type xml serialization:\n\n*file1.txt:*\n\n```text\nhello world!!!\n```\n\n*file2.txt:*\n\n```text\ntest\n```\n\n*main.py:*\n\n```python\nimport base64\nfrom typing import Any, List\nfrom pydantic_xml import BaseXmlModel, XmlEncoder, attr, element\n\n\nclass CustomXmlEncoder(XmlEncoder):\n    def encode(self, obj: Any) -> str:\n        if isinstance(obj, bytes):\n            return base64.b64encode(obj).decode()\n\n        return super().encode(obj)\n\n\nclass File(BaseXmlModel):\n    name: str = attr()\n    content: bytes = element()\n\n\nclass Files(BaseXmlModel, tag=\'files\'):\n    __root__: List[File] = element(tag=\'file\', default=[])\n\n\nfiles = Files()\nfor filename in [\'file1.txt\', \'file2.txt\']:\n    with open(filename, \'rb\') as f:\n        content = f.read()\n\n    files.__root__.append(File(name=filename, content=content))\n\nxml = files.to_xml(encoder=CustomXmlEncoder(), pretty_print=True, encoding=\'UTF-8\', standalone=True)\nprint(xml.decode())\n```\n\n*output:*\n\n```xml\n<?xml version=\'1.0\' encoding=\'UTF-8\' standalone=\'yes\'?>\n<files>\n  <file name="file1.txt">\n    <content>aGVsbG8gd29ybGQhISE=</content>\n  </file>\n  <file name="file2.txt">\n    <content>dGVzdA==</content>\n  </file>\n</files>\n```\n\n\n### Data modeling\n\nThis section provides examples of the most common use-cases with xml and json outputs.\n\n#### Primitive types:\n\n```python\nclass Company(BaseXmlModel):\n    class CompanyType(str, Enum):\n        PRIVATE = \'Private\'\n        PUBLIC = \'Public\'\n\n    trade_name: str = attr(name=\'trade-name\')\n    type: CompanyType = attr()  # field name is used as an attribute name by default\n    founded: Optional[date] = element(tag=\'Founded\')\n```\n\n```xml\n<Company trade-name="SpaceX" type="Private">\n    <Founded>2002-03-14</Founded>\n</Company>\n```\n\n```json\n{\n    "trade_name": "SpaceX",\n    "type": "Private",\n    "founded": "2002-03-14"\n}\n```\n--------------------------------------------------------------------------------\n\n```python\nclass Product(BaseXmlModel):\n    status: Literal[\'running\', \'development\'] = attr()\n    launched: Optional[int] = attr()\n    title: str  # primitive types are extracted from the element text by default\n```\n\n```xml\n<Product status="running" launched="2019">Starlink</Product>\n```\n\n```json\n{\n    "status": "running",\n    "launched": 2019,\n    "title": "Starlink"\n}\n```\n--------------------------------------------------------------------------------\n\n\n#### Sub-models:\n\n```python\nclass Headquarters(BaseXmlModel):\n    country: str = element()\n    state: str = element()\n    city: str = element()\n\n\nclass Product(BaseXmlModel):\n    status: Literal[\'running\', \'development\'] = attr()\n    launched: Optional[int] = attr()\n    title: str\n\n\nclass Company(BaseXmlModel):\n    headquarters: Headquarters  # models are extracted from the element by default\n    products: Tuple[Product, ...] = element(tag=\'product\')\n```\n\n```xml\n<Company>\n    <headquarters>\n        <country>US</country>\n        <state>California</state>\n        <city>Hawthorne</city>\n    </headquarters>\n\n    <product status="running" launched="2013">Several launch vehicles</product>\n    <product status="running" launched="2019">Starlink</product>\n    <product status="development">Starship</product>\n</Company>\n```\n\n```json\n{\n    "headquarters": {\n        "country": "US",\n        "state": "California",\n        "city": "Hawthorne"\n    },\n    "products": [\n        {\n            "status": "running",\n            "launched": 2013,\n            "title": "Several launch vehicles"\n        },\n        {\n            "status": "running",\n            "launched": 2019,\n            "title": "Starlink"\n        },\n        {\n            "status": "development",\n            "launched": null,\n            "title": "Starship"\n        }\n    ]\n}\n```\n--------------------------------------------------------------------------------\n\n```python\nclass Founded(BaseXmlModel):\n    __root__: datetime.date\n\nclass Company(BaseXmlModel):\n    founded: Founded = element(tag=\'Founded\')\n```\n\n```xml\n<Company>\n    <Founded>2002-03-14</Founded>\n</Company>\n```\n\n```json\n{\n    "founded": "2002-03-14"\n}\n```\n--------------------------------------------------------------------------------\n\n\n#### Mappings:\n\n```python\nclass Company(BaseXmlModel):\n    founder: Dict[str, str] = element(tag=\'Founder\')\n```\n\n```xml\n<Company>\n    <Founder name="Elon" surname="Musk"/>\n</Company>\n```\n\n```json\n{\n    "founder": {\n        "name": "Elon",\n        "surname": "Musk"\n    }\n}\n```\n--------------------------------------------------------------------------------\n\n```python\nclass Founder(BaseXmlModel):\n    __root__: Dict[str, str]  # mappings are extracted from attributes by default\n```\n\n```xml\n<Founder name="Elon" surname="Musk"/>\n```\n\n```json\n{\n    "name": "Elon",\n    "surname": "Musk"\n}\n```\n--------------------------------------------------------------------------------\n\n\n#### Homogeneous collections:\n\n```python\nclass Industries(BaseXmlModel):\n    __root__: Set[str] = element(tag=\'industry\')\n\n\nclass Social(BaseXmlModel):\n    type: str = attr()\n    url: str\n\n\nclass Product(BaseXmlModel):\n    status: Literal[\'running\', \'development\'] = attr()\n    launched: Optional[int] = attr()\n    title: str\n\n\nclass Company(BaseXmlModel):\n    industries: Industries = element()\n    socials: List[Social] = element(tag=\'social\', default_factory=list)\n    products: Tuple[Product, ...] = element(tag=\'product\')\n```\n\n```xml\n<Company>\n    <industries>\n        <industry>space</industry>\n        <industry>communications</industry>\n    </industries>\n\n    <social type="linkedin">https://www.linkedin.com/company/spacex</social>\n    <social type="twitter">https://twitter.com/spacex</social>\n    <social type="youtube">https://www.youtube.com/spacex</social>\n\n    <product status="running" launched="2013">Several launch vehicles</product>\n    <product status="running" launched="2019">Starlink</product>\n    <product status="development">Starship</product>\n</Company>\n```\n\n```json\n{\n    "industries": [\n        "space",\n        "communications"\n    ],\n    "socials": [\n        {\n            "type": "linkedin",\n            "url": "https://www.linkedin.com/company/spacex"\n        },\n        {\n            "type": "twitter",\n            "url": "https://twitter.com/spacex"\n        },\n        {\n            "type": "youtube",\n            "url": "https://www.youtube.com/spacex"\n        }\n    ],\n    "products": [\n        {\n            "status": "running",\n            "launched": 2013,\n            "title": "Several launch vehicles"\n        },\n        {\n            "status": "running",\n            "launched": 2019,\n            "title": "Starlink"\n        },\n        {\n            "status": "development",\n            "launched": null,\n            "title": "Starship"\n        }\n    ]\n}\n```\n--------------------------------------------------------------------------------\n\n\n#### Heterogeneous collections:\n\n```python\nclass Person(BaseXmlModel):\n    name: str = attr()\n\n\nclass CEO(Person):\n    position: Literal[\'CEO\'] = attr()\n\n\nclass CTO(Person):\n    position: Literal[\'CTO\'] = attr()\n\n\nclass COO(Person):\n    position: Literal[\'COO\'] = attr()\n\n\nclass Company(BaseXmlModel):\n    key_people: Tuple[CEO, CTO, COO] = element(tag=\'person\')\n```\n\n```xml\n<Company>\n    <person position="CEO" name="Elon Musk"/>\n    <person position="CTO" name="Elon Musk"/>\n    <person position="COO" name="Gwynne Shotwell"/>\n</Company>\n```\n\n```json\n{\n    "key_people": [\n        {\n            "name": "Elon Musk",\n            "position": "CEO"\n        },\n        {\n            "name": "Elon Musk",\n            "position": "CTO"\n        },\n        {\n            "name": "Gwynne Shotwell",\n            "position": "COO"\n        }\n    ]\n}\n```\n--------------------------------------------------------------------------------\n\n\n#### Wrapped entities:\n\n```python\nclass Company(BaseXmlModel):\n    hq_country: str = wrapped(\'headquarters/country\')\n    hq_state: str = wrapped(\'headquarters/state\')\n    hq_city: str = wrapped(\'headquarters/city\')\n```\n\n```xml\n<Company>\n    <headquarters>\n        <country>US</country>\n        <state>California</state>\n        <city>Hawthorne</city>\n    </headquarters>\n</Company>\n```\n\n```json\n{\n    "hq_country": "US",\n    "hq_state": "California",\n    "hq_city": "Hawthorne"\n}\n```\n--------------------------------------------------------------------------------\n\n\n```python\nclass Social(BaseXmlModel):\n    type: str = attr()\n    url: str\n\n\nclass Company(BaseXmlModel):\n    socials: List[Social] = wrapped(\'contacts/socials\', element(tag=\'social\', default_factory=list))\n```\n\n```xml\n<Company>\n    <contacts>\n        <socials>\n            <social type="linkedin">https://www.linkedin.com/company/spacex</social>\n            <social type="twitter">https://twitter.com/spacex</social>\n            <social type="youtube">https://www.youtube.com/spacex</social>\n        </socials>\n    </contacts>\n</Company>\n```\n\n```json\n{\n    "socials": [\n        {\n            "type": "linkedin",\n            "url": "https://www.linkedin.com/company/spacex"\n        },\n        {\n            "type": "twitter",\n            "url": "https://twitter.com/spacex"\n        },\n        {\n            "type": "youtube",\n            "url": "https://www.youtube.com/spacex"\n        }\n    ]\n}\n```\n--------------------------------------------------------------------------------\n\n```python\nclass Company(BaseXmlModel):\n    persons: List[Dict[str, str]] = wrapped(\'key-people\', element(tag="person"))\n```\n\n```xml\n<Company>\n    <key-people>\n        <person position="CEO" name="Elon Musk"/>\n        <person position="CTO" name="Elon Musk"/>\n        <person position="COO" name="Gwynne Shotwell"/>\n    </key-people>\n</Company>\n```\n\n```json\n{\n    "persons": [\n        {\n            "position": "CEO",\n            "name": "Elon Musk"\n        },\n        {\n            "position": "CTO",\n            "name": "Elon Musk"\n        },\n        {\n            "position": "COO",\n            "name": "Gwynne Shotwell"\n        }\n    ]\n}\n```\n--------------------------------------------------------------------------------\n',
    'author': 'Dmitry Pershin',
    'author_email': 'dapper1291@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/dapper91/pydantic-xml',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, ClassVar, Dict, Optional, Tuple, Type, Union

import pydantic as pd
import pydantic.fields
import pydantic.generics

from . import config, errors, serializers
from .backend import etree
from .utils import NsMap, register_nsmap


class XmlEntityInfo(pd.fields.FieldInfo):
    """
    Field xml meta-information.
    """


class XmlAttributeInfo(XmlEntityInfo):
    """
    Field xml attribute meta-information.

    :param tag: attribute name
    :param ns: attribute xml namespace
    :param kwargs: pydantic field arguments
    """

    def __init__(
            self,
            name: Optional[str] = None,
            ns: Optional[str] = None,
            **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._name = name
        self._ns = ns

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def ns(self) -> Optional[str]:
        return self._ns


class XmlElementInfo(XmlEntityInfo):
    """
    Field xml element meta-information.

    :param tag: element tag
    :param ns: element xml namespace
    :param nsmap: element xml namespace map
    :param kwargs: pydantic field arguments
    """

    def __init__(
            self,
            tag: Optional[str] = None,
            ns: Optional[str] = None,
            nsmap: Optional[NsMap] = None,
            **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._tag = tag
        self._ns = ns
        self._nsmap = nsmap

        if config.REGISTER_NS_PREFIXES and nsmap:
            register_nsmap(nsmap)

    @property
    def tag(self) -> Optional[str]:
        return self._tag

    @property
    def ns(self) -> Optional[str]:
        return self._ns

    @property
    def nsmap(self) -> Optional[NsMap]:
        return self._nsmap


class XmlWrapperInfo(XmlEntityInfo):
    """
    Field xml wrapper meta-information.

    :param entity: wrapped entity
    :param tag: element tag
    :param ns: element xml namespace
    :param nsmap: element xml namespace map
    :param kwargs: pydantic field arguments
    """

    def __init__(
            self,
            path: str,
            entity: Optional[XmlEntityInfo] = None,
            ns: Optional[str] = None,
            nsmap: Optional[NsMap] = None,
            **kwargs: Any,
    ):
        if entity is not None:
            # copy arguments from the wrapped entity to let pydantic know how to process the field
            for entity_field_name in entity.__slots__:
                kwargs[entity_field_name] = getattr(entity, entity_field_name)

        super().__init__(**kwargs)
        self._entity = entity
        self._path = path
        self._ns = ns
        self._nsmap = nsmap

        if config.REGISTER_NS_PREFIXES and nsmap:
            register_nsmap(nsmap)

    @property
    def entity(self) -> Optional[XmlEntityInfo]:
        return self._entity

    @property
    def path(self) -> str:
        return self._path

    @property
    def ns(self) -> Optional[str]:
        return self._ns

    @property
    def nsmap(self) -> Optional[NsMap]:
        return self._nsmap


def attr(**kwargs: Any) -> XmlAttributeInfo:
    """
    Marks a pydantic field as an xml attribute.
    """

    return XmlAttributeInfo(**kwargs)


def element(**kwargs: Any) -> XmlElementInfo:
    """
    Marks a pydantic field as an xml element.
    """

    return XmlElementInfo(**kwargs)


def wrapped(*args: Any, **kwargs: Any) -> XmlWrapperInfo:
    """
    Marks a pydantic field as a wrapped xml entity.
    """

    return XmlWrapperInfo(*args, **kwargs)


class XmlModelMeta(pd.main.ModelMetaclass):

    __is_base_model_defined__ = False

    def __new__(mcls, name: str, bases: Tuple[type], namespace: Dict[str, Any], **kwargs: Any) -> Type['BaseXmlModel']:
        cls = super().__new__(mcls, name, bases, namespace, **kwargs)
        if mcls.__is_base_model_defined__:
            cls.__init_serializer__()
        else:
            mcls.__is_base_model_defined__ = True

        return cls


class BaseXmlModel(pd.BaseModel, metaclass=XmlModelMeta):
    """
    Base pydantic-xml model.
    """

    __xml_tag__: ClassVar[Optional[str]]
    __xml_ns__: ClassVar[Optional[str]]
    __xml_nsmap__: ClassVar[Optional[NsMap]]
    __xml_ns_attrs__: ClassVar[bool]
    __xml_serializer__: ClassVar[Optional[serializers.ModelSerializerFactory.RootSerializer]]

    def __init_subclass__(
            cls,
            *args: Any,
            tag: Optional[str] = None,
            ns: Optional[str] = None,
            nsmap: Optional[NsMap] = None,
            ns_attrs: bool = False,
            **kwargs: Any,
    ):
        """
        Initializes a subclass.

        :param tag: element tag
        :param ns: element namespace
        :param nsmap: element namespace map
        :param ns_attrs: use namespaced attributes
        """

        super().__init_subclass__(*args, **kwargs)

        cls.__xml_tag__ = tag
        cls.__xml_ns__ = ns
        cls.__xml_nsmap__ = nsmap
        cls.__xml_ns_attrs__ = ns_attrs

    @classmethod
    def __init_serializer__(cls) -> None:
        if config.REGISTER_NS_PREFIXES and cls.__xml_nsmap__:
            register_nsmap(cls.__xml_nsmap__)

        cls.__xml_serializer__ = serializers.ModelSerializerFactory.from_model(cls)

    @classmethod
    def from_xml_tree(cls, root: etree.Element) -> Optional['BaseXmlModel']:
        """
        Deserializes an xml element tree to an object of `cls` type.

        :param root: xml element to deserialize the object from
        :return: deserialized object
        """

        assert cls.__xml_serializer__ is not None, "model is partially initialized"

        if root.tag == cls.__xml_serializer__.element_name:
            obj = cls.__xml_serializer__.deserialize(root)
            return cls.parse_obj(obj)
        else:
            return None

    @classmethod
    def from_xml(cls, source: Union[str, bytes]) -> Optional['BaseXmlModel']:
        """
        Deserializes an xml string to an object of `cls` type.

        :param source: xml string
        :return: deserialized object
        """

        return cls.from_xml_tree(etree.fromstring(source))

    def to_xml_tree(
            self,
            *,
            encoder: Optional[serializers.XmlEncoder] = None,
            skip_empty: bool = False,
    ) -> etree.Element:
        """
        Serializes the object to an xml tree.

        :param encoder: xml type encoder
        :param skip_empty: skip empty elements (elements without sub-elements, attributes and text, Nones)
        :return: object xml representation
        """

        encoder = encoder or serializers.DEFAULT_ENCODER

        assert self.__xml_serializer__ is not None, "model is partially initialized"
        root = self.__xml_serializer__.serialize(None, self, encoder=encoder, skip_empty=skip_empty)
        assert root is not None

        return root

    def to_xml(
            self,
            *,
            encoder: Optional[serializers.XmlEncoder] = None,
            skip_empty: bool = False,
            **kwargs: Any,
    ) -> Union[str, bytes]:
        """
        Serializes the object to an xml string.

        :param encoder: xml type encoder
        :param skip_empty: skip empty elements (elements without sub-elements, attributes and text, Nones)
        :param kwargs: additional xml serialization arguments
        :return: object xml representation
        """

        return etree.tostring(self.to_xml_tree(encoder=encoder, skip_empty=skip_empty), **kwargs)


class BaseGenericXmlModel(BaseXmlModel, pd.generics.GenericModel):
    """
    Base pydantic-xml generic model.
    """

    def __class_getitem__(cls, params: Union[Type[Any], Tuple[Type[Any], ...]]) -> Type[Any]:
        model = super().__class_getitem__(params)
        model.__xml_tag__ = cls.__xml_tag__
        model.__xml_ns__ = cls.__xml_ns__
        model.__xml_nsmap__ = cls.__xml_nsmap__
        model.__xml_ns_attrs__ = cls.__xml_ns_attrs__
        model.__init_serializer__()

        return model

    @classmethod
    def __init_serializer__(cls) -> None:
        # checks that the model is not generic
        if not getattr(cls, '__concrete__', True):
            cls.__xml_serializer__ = None
        else:
            super().__init_serializer__()

    @classmethod
    def from_xml_tree(cls, root: etree.Element) -> Optional['BaseXmlModel']:
        """
        Deserializes an xml element tree to an object of `cls` type.

        :param root: xml element to deserialize the object from
        :return: deserialized object
        """

        if cls.__xml_serializer__ is None:
            raise errors.ModelError(f"{cls.__name__} model is generic")

        return super().from_xml_tree(root)



import xml.etree.ElementTree as etree
from typing import Any, Dict, Optional

__all__ = (
    'etree',
    'create_element',
)


def create_element(
        tag: str,
        attrib: Optional[Dict[str, Any]] = None,
        nsmap: Optional[Dict[str, str]] = None,  # not supported by `xml.etree`
) -> etree.Element:
    return etree.Element(tag, attrib=attrib or {})
